#!/usr/bin/env node
'use strict';


const process = require("process"),
    net = require("net"),
    fs = require('fs'),
    path = require('path'),
    pathInfo = require('./pathInfo'),
    logFun = require('./log'),
    argv= process.argv,
    log = logFun(),
    gameLog = logFun("GameLog"),
    pkgObj = require("./package.json"),
    style = require("./style"),
    createTaskId = require("./createTaskId");


class LuaMon {
    constructor(option){
        const _ts = this;
        _ts.option = {
            port:43996,
            hostname:"127.0.0.1"
        };
        option = option || {};
        for(let key in option){
            _ts.option[key] = option[key];
        };

        if(_ts.option.params["--version"] || _ts.option.params["-v"]){
            console.log(`${pkgObj.version}`);
            return;
        };
        if(_ts.option.params["--help"] || _ts.option.params["-h"]){
            console.log(`LuaMon Version ${pkgObj.version}`);
            console.log(`luamon ip                ${style('gray','自定义服务器地址')}`);
            console.log(`luamon :port             ${style('gray','自定义连接端口')}`);
            console.log(`luamon ip:port           ${style('gray','自定义服务器地址及端口')}`);
            console.log(`luamon --test            ${style('gray','仅对 *@test.lua 的文件进行监听')}`);
            console.log(`luamon --hotload         ${style('gray','启用热加载')}`);
            console.log(`luamon --log             ${style('gray','输出游戏日志。默认全部日志，可传入数值指定只输出某种类型的日志（0 错误日志、1 警告日志、2正常日志）')}`);
            console.log(`luamon --version   -v    ${style('gray','查看当前luamon版本')}`);
            console.log(`luamon --help      -h    ${style('gray','查看帮助')}`);
            return;
        };
        
        _ts.blockSize = 1024 * 512;                                                                 // win10 64下测试最大只能 1024 * 1023，否则发送中断
        _ts.sendQueue = [];
        _ts.isConnected = false;
        _ts.watchPath = process.cwd() || _ts.option.path || __dirname;
        _ts.serverAdrStr = `${_ts.option.hostname}:${_ts.option.port}`;
        _ts.temp = {};
        _ts.isReConnect = false;
        _ts.serverStatus = "idle";                                                                  // 服务器状态（connection连接、idle闲置、occupy占用、disconnect断开）

        log.print(`监听路径 -> ${_ts.watchPath}`);
        
        fs.watch(_ts.watchPath,{recursive:true},(...arg)=>{                                         // 监听
            _ts.watchFiles(...arg);
        });

        _ts.init();                                                                                 // 初始化连接
    }
    watchFiles(eventType,fileName){
        const _ts = this;
        if(typeof fileName !== "string"){                                                           // 处理部分临时文件生成又删除可能获取不到文件名出现的错误
            return;
        };
        let filePath = path.join(_ts.watchPath,fileName),
            filePathInfo = pathInfo(filePath),
            re = /@test\.lua$/i;
        if(filePathInfo.type === 'file' && filePathInfo.extension === 'lua'){
            if(_ts.option.params["--test"] && !re.test(fileName)){
                return;
            };
            clearTimeout(_ts.temp[filePath]);                                                       // 防抖处理
            _ts.temp[filePath] = setTimeout(()=>{
                if(filePathInfo.isExist){
                    _ts.sendFile(filePath);
                }else{                                                                              // 删除文件操作
                    _ts.send(_ts.createData(
                        'removeFile',
                        null,
                        null,
                        null,
                        filePath
                    ));
                };
                log.print(`文件改变 -> ${filePath} -> ${eventType}`);
            },200);
        };
    }

    triggerQueue(){
        const _ts = this;
        
        if(_ts.serverStatus !== "idle"){
            log.error("服务器不是闲置状态");
            return;
        };

        if(_ts.sendQueue.length == 0){
            return;
        };
        
        let item = _ts.sendQueue.shift();                                                           // 触发队列
        _ts.send(_ts.createData(
            item.type,
            item.id,
            item.index,
            item.total,
            item.path,
            item.data
        ));
    }
    
    init(){
        const _ts = this;
        _ts.socket  = new net.Socket();
        _ts.socket.setEncoding("utf8");
        _ts.socket.on('data', (...args)=>{
            _ts.receive(...args);
        });
        _ts.socket.on('error', (...args)=>{
            _ts.error(...args);
        });
        _ts.socket.on('close',(...args)=>{
            _ts.close(...args);
        });
        _ts.socket.on('end',(...args)=>{
            _ts.end(...args);
        });
        let port = this.option.port;
        let hostname = this.option.hostname;
        _ts.socket.connect(port,hostname,()=>{
            log.print(`连接成功 -> ${_ts.serverAdrStr}`);

            // 发送日志输出方式
            if(_ts.option.params["--log"]){
                let val = "--log";
                if(_ts.option.params["--log"] !== true){
                    val += `=${_ts.option.params["--log"]}`;
                };
                _ts.sendSetting(val);
            };

            // 发送是否热加载
            if(_ts.option.params["--hotload"]){
                _ts.sendSetting("--hotload");
            };
        });
    }
    end(){
        const _ts = this;
        log.print(`连接断开 -> ${_ts.serverAdrStr}`);
        _ts.sendQueue = [];                                                                         // 意外断开时清空队列
    }

    // 发送文件
    sendFile(filePath){
        const _ts = this;
        let fileBuffer = fs.readFileSync(filePath),
            blockLen = fileBuffer.length / _ts.blockSize,
            taskId = createTaskId();

        // 数据分块缓存到任务队列
        for(let i=0; i<blockLen; i++){
            let start = i * _ts.blockSize,
                end = i >= blockLen ? fileBuffer.length : start + _ts.blockSize;
            _ts.sendQueue.push({
                type:"sendFile",
                id:taskId,
                index:i,
                total:fileBuffer.length,
                block:end - start,
                path:filePath,
                data:fileBuffer.slice(start,end)
            });
        };

        // 尝试触发队列
        _ts.triggerQueue();
    }

    // 发送数据方法，不建议直接调用
    send(data){
        this.socket.write(data);
    }

    // 发送设置指令
    sendSetting(setting){
        this.send(this.createData(
            'setting',
            null,
            null,
            null,
            null,
            Buffer.from(setting)
        ));
    }

    // 创建发送数据
    createData(type,id,index,total,path,data){
        if(!type){
            throw new Error("数据类型不能为空");
        };
        data = data || Buffer.alloc(0,0);
        id = id || createTaskId();
        index = index || "0";
        index = index.toString();
        total = total || data.length.toString();
        total = total.toString();
        path = path || "";
        
        const hedaer = Buffer.alloc(40 + 10 + 10 + 10 + 10 + 260,0),
            block = data.length.toString();

        hedaer.write(id,0);                                                                         // id，例如：1620282871025_23217769489186484，第 0-40 位，40位长
        hedaer.write(type,40);                                                                      // type，用于声明通信类型，第 40-50 位，10位长
        hedaer.write(index,50);                                                                     // index，用于记录文件的插入位置，第 50-60 位，10位长
        hedaer.write(total,60);                                                                     // total，用于记录文件总大小，第 60-70 位，10位长
        hedaer.write(block,70);                                                                     // block，记录当前块大小，第 70-80 位，10位长
        hedaer.write(path,80);                                                                      // path，用于记录文件路径，第 80-340 位，260位长
        return Buffer.concat([hedaer,data]);
    }

    receive(chunk){
        const _ts = this;
        let type = chunk.slice(0,3),
            content = chunk.slice(3);
        switch (type) {
            case "00@":                                                                             // 脚本运行错误
                log.error(content);
            break;
            case "01@":                                                                             // 脚本运行警告
                log.warning(content);
            break;
            case "02@":                                                                             // 脚本普通提示
                log.info(content);
            break;
            case "03@":                                                                             // 脚本成功运行
                log.success(content);
            break;
            case "10@":                                                                             // 游戏错误日志
                gameLog.error(content);
            break;
            case "11@":                                                                             // 游戏警告日志
                gameLog.warning(content);
            break;
            case "12@":                                                                             // 游戏普通日志
                gameLog.info(content);
            break;
            case "90@":                                                                             // 指令
                let obj = JSON.parse(content);
                switch (obj.command) {
                    case "nextTask":
                        _ts.triggerQueue();
                    break;
                    default:
                        log.error(`未处理的指令 -> ${content}`);
                    break;
                };
            break;
            default:                                                                                // 其它
                log.info(content);
            break;
        };
    }
    error(err){
        const _ts = this;
        if(!_ts.isReConnect){
            log.print(`连接出错 -> ${err.message}`);
        };
    }
    close(isClose){
        const _ts = this;
        let msg = isClose ? `尝试重连` : `连接关闭`;
        if(!_ts.isReConnect){
            log.print(`${msg} -> ${_ts.serverAdrStr}`);
        };
        _ts.isReConnect = isClose;
        _ts.init();                                                                                 // 关闭之后尝试再连接
    }
}

const option = (()=>{                                                                               // 跟据进程初始化选项
        let result = {
                params:{}
            },
            args = argv.slice(2);
        let res = {
            ip:/^(\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])(\.(\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])){3}$/,
            ipPort:/^(\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])(\.(\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])){3}:([0-9]|[1-9]\d|[1-9]\d{2}|[1-9]\d{3}|[1-5]\d{4}|6[0-4]\d{3}|65[0-4]\d{2}|655[0-2]\d|6553[0-5])$/,
            port:/^:([0-9]|[1-9]\d|[1-9]\d{2}|[1-9]\d{3}|[1-5]\d{4}|6[0-4]\d{3}|65[0-4]\d{2}|655[0-2]\d|6553[0-5])$/,
            param:/^(-{1,2}\w{1,})$/
        },

        // 检查是否为 cli 参数
        isCliParam = item => {
            for(let key in res){
                let re = res[key];
                if(re.test(item)){
                    return true;
                };
            };
            return false;
        };
        
        for(let i=0,len=args.length; i<len; i++){
            let item = args[i];

            b:for(let key in res){
                let re = res[key];
                if(re.test(item)){
                    switch (key) {
                        case "ip":
                            result.hostname = item;
                        break;
                        case "ipPort":
                            let arrItem = item.split(":");
                            result.hostname = arrItem[0];
                            result.port = arrItem[1];
                        break;
                        case "port":
                            result.port = item.slice(1);
                        break
                        case "param":
                            // 如果有手动赋值由采用传入的值，否则默认为true
                            let nextItem = args[i+1];
                            result.params[item.toLocaleLowerCase()] = nextItem !== undefined && !isCliParam(nextItem) ? nextItem : true;
                        break;
                        default:
                        break;
                    };
                    break b;
                };
            };
        };
        return result;
    })(),
    luamon = new LuaMon(option);